@@ -8,10 +8,12 @@
 
 #include "resolve-names-utils.h"
 #include "flang/Common/Fortran-features.h"
+#include "flang/Common/Fortran.h"
 #include "flang/Common/idioms.h"
 #include "flang/Common/indirection.h"
 #include "flang/Evaluate/fold.h"
 #include "flang/Evaluate/tools.h"
+#include "flang/Evaluate/traverse.h"
 #include "flang/Evaluate/type.h"
 #include "flang/Parser/char-block.h"
 #include "flang/Parser/parse-tree.h"
@@ -43,10 +45,10 @@ Symbol &Resolve(const parser::Name &name, Symbol &symbol) {
   return *Resolve(name, &symbol);
 }
 
-parser::MessageFixedText WithIsFatal(
-    const parser::MessageFixedText &msg, bool isFatal) {
+parser::MessageFixedText WithSeverity(
+    const parser::MessageFixedText &msg, parser::Severity severity) {
   return parser::MessageFixedText{
-      msg.text().begin(), msg.text().size(), isFatal};
+      msg.text().begin(), msg.text().size(), severity};
 }
 
 bool IsIntrinsicOperator(
@@ -54,13 +56,13 @@ bool IsIntrinsicOperator(
   std::string str{name.ToString()};
   for (int i{0}; i != common::LogicalOperator_enumSize; ++i) {
     auto names{context.languageFeatures().GetNames(LogicalOperator{i})};
-    if (std::find(names.begin(), names.end(), str) != names.end()) {
+    if (llvm::is_contained(names, str)) {
       return true;
     }
   }
   for (int i{0}; i != common::RelationalOperator_enumSize; ++i) {
     auto names{context.languageFeatures().GetNames(RelationalOperator{i})};
-    if (std::find(names.begin(), names.end(), str) != names.end()) {
+    if (llvm::is_contained(names, str)) {
       return true;
     }
   }
@@ -84,13 +86,13 @@ std::forward_list<std::string> GetAllNames(
       name.ToString().rfind(std::string{operatorPrefix}, 0) == 0) {
     for (int i{0}; i != common::LogicalOperator_enumSize; ++i) {
       auto names{GetOperatorNames(context, LogicalOperator{i})};
-      if (std::find(names.begin(), names.end(), str) != names.end()) {
+      if (llvm::is_contained(names, str)) {
         return names;
       }
     }
     for (int i{0}; i != common::RelationalOperator_enumSize; ++i) {
       auto names{GetOperatorNames(context, RelationalOperator{i})};
-      if (std::find(names.begin(), names.end(), str) != names.end()) {
+      if (llvm::is_contained(names, str)) {
         return names;
       }
     }
@@ -125,7 +127,7 @@ void GenericSpecInfo::Analyze(const parser::DefinedOpName &name) {
 
 void GenericSpecInfo::Analyze(const parser::GenericSpec &x) {
   symbolName_ = x.source;
-  kind_ = std::visit(
+  kind_ = common::visit(
       common::visitors{
           [&](const parser::Name &y) -> GenericKind {
             parseName_ = &y;
@@ -133,7 +135,7 @@ void GenericSpecInfo::Analyze(const parser::GenericSpec &x) {
             return GenericKind::OtherKind::Name;
           },
           [&](const parser::DefinedOperator &y) {
-            return std::visit(
+            return common::visit(
                 common::visitors{
                     [&](const parser::DefinedOpName &z) -> GenericKind {
                       Analyze(z);
@@ -264,19 +266,20 @@ ArraySpec AnalyzeCoarraySpec(
 }
 
 ArraySpec ArraySpecAnalyzer::Analyze(const parser::ComponentArraySpec &x) {
-  std::visit([this](const auto &y) { Analyze(y); }, x.u);
+  common::visit([this](const auto &y) { Analyze(y); }, x.u);
   CHECK(!arraySpec_.empty());
   return arraySpec_;
 }
 ArraySpec ArraySpecAnalyzer::Analyze(const parser::ArraySpec &x) {
-  std::visit(common::visitors{
-                 [&](const parser::AssumedSizeSpec &y) {
-                   Analyze(std::get<std::list<parser::ExplicitShapeSpec>>(y.t));
-                   Analyze(std::get<parser::AssumedImpliedSpec>(y.t));
-                 },
-                 [&](const parser::ImpliedShapeSpec &y) { Analyze(y.v); },
-                 [&](const auto &y) { Analyze(y); },
-             },
+  common::visit(common::visitors{
+                    [&](const parser::AssumedSizeSpec &y) {
+                      Analyze(
+                          std::get<std::list<parser::ExplicitShapeSpec>>(y.t));
+                      Analyze(std::get<parser::AssumedImpliedSpec>(y.t));
+                    },
+                    [&](const parser::ImpliedShapeSpec &y) { Analyze(y.v); },
+                    [&](const auto &y) { Analyze(y); },
+                },
       x.u);
   CHECK(!arraySpec_.empty());
   return arraySpec_;
@@ -288,7 +291,7 @@ ArraySpec ArraySpecAnalyzer::AnalyzeDeferredShapeSpecList(
   return arraySpec_;
 }
 ArraySpec ArraySpecAnalyzer::Analyze(const parser::CoarraySpec &x) {
-  std::visit(
+  common::visit(
       common::visitors{
           [&](const parser::DeferredCoshapeSpecList &y) { MakeDeferred(y.v); },
           [&](const parser::ExplicitCoshapeSpec &y) {
@@ -303,7 +306,7 @@ ArraySpec ArraySpecAnalyzer::Analyze(const parser::CoarraySpec &x) {
 }
 
 void ArraySpecAnalyzer::Analyze(const parser::AssumedShapeSpec &x) {
-  arraySpec_.push_back(ShapeSpec::MakeAssumed(GetBound(x.v)));
+  arraySpec_.push_back(ShapeSpec::MakeAssumedShape(GetBound(x.v)));
 }
 void ArraySpecAnalyzer::Analyze(const parser::ExplicitShapeSpec &x) {
   MakeExplicit(std::get<std::optional<parser::SpecificationExpr>>(x.t),
@@ -412,47 +415,59 @@ void EquivalenceSets::FinishSet(const parser::CharBlock &source) {
   currSet_.clear();
 }
 
-// Report an error if sym1 and sym2 cannot be in the same equivalence set.
+// Report an error or warning if sym1 and sym2 cannot be in the same equivalence
+// set.
 bool EquivalenceSets::CheckCanEquivalence(
     const parser::CharBlock &source, const Symbol &sym1, const Symbol &sym2) {
   std::optional<parser::MessageFixedText> msg;
   const DeclTypeSpec *type1{sym1.GetType()};
   const DeclTypeSpec *type2{sym2.GetType()};
-  bool isNum1{IsNumericSequenceType(type1)};
-  bool isNum2{IsNumericSequenceType(type2)};
+  bool isDefaultNum1{IsDefaultNumericSequenceType(type1)};
+  bool isAnyNum1{IsAnyNumericSequenceType(type1)};
+  bool isDefaultNum2{IsDefaultNumericSequenceType(type2)};
+  bool isAnyNum2{IsAnyNumericSequenceType(type2)};
   bool isChar1{IsCharacterSequenceType(type1)};
   bool isChar2{IsCharacterSequenceType(type2)};
   if (sym1.attrs().test(Attr::PROTECTED) &&
       !sym2.attrs().test(Attr::PROTECTED)) { // C8114
     msg = "Equivalence set cannot contain '%s'"
           " with PROTECTED attribute and '%s' without"_err_en_US;
-  } else if (isNum1) {
+  } else if ((isDefaultNum1 && isDefaultNum2) || (isChar1 && isChar2)) {
+    // ok & standard conforming
+  } else if (!(isAnyNum1 || isChar1) &&
+      !(isAnyNum2 || isChar2)) { // C8110 - C8113
+    if (AreTkCompatibleTypes(type1, type2)) {
+      if (context_.ShouldWarn(LanguageFeature::EquivalenceSameNonSequence)) {
+        msg =
+            "nonstandard: Equivalence set contains '%s' and '%s' with same "
+            "type that is neither numeric nor character sequence type"_port_en_US;
+      }
+    } else {
+      msg = "Equivalence set cannot contain '%s' and '%s' with distinct types "
+            "that are not both numeric or character sequence types"_err_en_US;
+    }
+  } else if (isAnyNum1) {
     if (isChar2) {
       if (context_.ShouldWarn(
               LanguageFeature::EquivalenceNumericWithCharacter)) {
-        msg = "Equivalence set contains '%s' that is numeric sequence "
-              "type and '%s' that is character"_en_US;
+        msg = "nonstandard: Equivalence set contains '%s' that is numeric "
+              "sequence type and '%s' that is character"_port_en_US;
       }
-    } else if (!isNum2) { // C8110
-      msg = "Equivalence set cannot contain '%s'"
-            " that is numeric sequence type and '%s' that is not"_err_en_US;
-    }
-  } else if (isChar1) {
-    if (isNum2) {
-      if (context_.ShouldWarn(
-              LanguageFeature::EquivalenceNumericWithCharacter)) {
-        msg = "Equivalence set contains '%s' that is character sequence "
-              "type and '%s' that is numeric"_en_US;
+    } else if (isAnyNum2 &&
+        context_.ShouldWarn(LanguageFeature::EquivalenceNonDefaultNumeric)) {
+      if (isDefaultNum1) {
+        msg =
+            "nonstandard: Equivalence set contains '%s' that is a default "
+            "numeric sequence type and '%s' that is numeric with non-default kind"_port_en_US;
+      } else if (!isDefaultNum2) {
+        msg = "nonstandard: Equivalence set contains '%s' and '%s' that are "
+              "numeric sequence types with non-default kinds"_port_en_US;
       }
-    } else if (!isChar2) { // C8111
-      msg = "Equivalence set cannot contain '%s'"
-            " that is character sequence type and '%s' that is not"_err_en_US;
     }
-  } else if (!isNum2 && !isChar2 && *type1 != *type2) { // C8112, C8113
-    msg = "Equivalence set cannot contain '%s' and '%s' with different types"
-          " that are neither numeric nor character sequence types"_err_en_US;
   }
-  if (msg) {
+  if (msg &&
+      (!context_.IsInModuleFile(source) ||
+          msg->severity() == parser::Severity::Error)) {
     context_.Say(source, std::move(*msg), sym1.name(), sym2.name());
     return false;
   }
@@ -484,7 +499,7 @@ const EquivalenceObject *EquivalenceSets::Find(
 }
 
 bool EquivalenceSets::CheckDesignator(const parser::Designator &designator) {
-  return std::visit(
+  return common::visit(
       common::visitors{
           [&](const parser::DataRef &x) {
             return CheckDataRef(designator.source, x);
@@ -509,7 +524,7 @@ bool EquivalenceSets::CheckDesignator(const parser::Designator &designator) {
 
 bool EquivalenceSets::CheckDataRef(
     const parser::CharBlock &source, const parser::DataRef &x) {
-  return std::visit(
+  return common::visit(
       common::visitors{
           [&](const parser::Name &name) { return CheckObject(name); },
           [&](const common::Indirection<parser::StructureComponent> &) {
@@ -521,7 +536,7 @@ bool EquivalenceSets::CheckDataRef(
           [&](const common::Indirection<parser::ArrayElement> &elem) {
             bool ok{CheckDataRef(source, elem.value().base)};
             for (const auto &subscript : elem.value().subscripts) {
-              ok &= std::visit(
+              ok &= common::visit(
                   common::visitors{
                       [&](const parser::SubscriptTriplet &) {
                         context_.Say(source, // C924, R872
@@ -562,7 +577,7 @@ bool EquivalenceSets::CheckObject(const parser::Name &name) {
     return false; // an error has already occurred
   }
   currObject_.symbol = name.symbol;
-  parser::MessageFixedText msg{"", 0};
+  parser::MessageFixedText msg;
   const Symbol &symbol{*name.symbol};
   if (symbol.owner().IsDerivedType()) { // C8107
     msg = "Derived type component '%s'"
@@ -605,7 +620,7 @@ bool EquivalenceSets::CheckObject(const parser::Name &name) {
         msg = "Nonsequence derived type object '%s'"
               " is not allowed in an equivalence set"_err_en_US;
       }
-    } else if (IsAutomaticObject(symbol)) {
+    } else if (IsAutomatic(symbol)) {
       msg = "Automatic object '%s'"
             " is not allowed in an equivalence set"_err_en_US;
     }
@@ -678,15 +693,14 @@ bool EquivalenceSets::IsCharacterSequenceType(const DeclTypeSpec *type) {
 // Numeric or logical type of default kind or DOUBLE PRECISION or DOUBLE COMPLEX
 bool EquivalenceSets::IsDefaultKindNumericType(const IntrinsicTypeSpec &type) {
   if (auto kind{evaluate::ToInt64(type.kind())}) {
-    auto category{type.category()};
-    auto defaultKind{context_.GetDefaultKind(category)};
-    switch (category) {
+    switch (type.category()) {
     case TypeCategory::Integer:
     case TypeCategory::Logical:
-      return *kind == defaultKind;
+      return *kind == context_.GetDefaultKind(TypeCategory::Integer);
     case TypeCategory::Real:
     case TypeCategory::Complex:
-      return *kind == defaultKind || *kind == context_.doublePrecisionKind();
+      return *kind == context_.GetDefaultKind(TypeCategory::Real) ||
+          *kind == context_.doublePrecisionKind();
     default:
       return false;
     }
@@ -694,12 +708,19 @@ bool EquivalenceSets::IsDefaultKindNumericType(const IntrinsicTypeSpec &type) {
   return false;
 }
 
-bool EquivalenceSets::IsNumericSequenceType(const DeclTypeSpec *type) {
+bool EquivalenceSets::IsDefaultNumericSequenceType(const DeclTypeSpec *type) {
   return IsSequenceType(type, [&](const IntrinsicTypeSpec &type) {
     return IsDefaultKindNumericType(type);
   });
 }
 
+bool EquivalenceSets::IsAnyNumericSequenceType(const DeclTypeSpec *type) {
+  return IsSequenceType(type, [&](const IntrinsicTypeSpec &type) {
+    return type.category() == TypeCategory::Logical ||
+        common::IsNumericTypeCategory(type.category());
+  });
+}
+
 // Is type an intrinsic type that satisfies predicate or a sequence type
 // whose components do.
 bool EquivalenceSets::IsSequenceType(const DeclTypeSpec *type,
@@ -722,4 +743,189 @@ bool EquivalenceSets::IsSequenceType(const DeclTypeSpec *type,
   }
 }
 
+// MapSubprogramToNewSymbols() relies on the following recursive symbol/scope
+// copying infrastructure to duplicate an interface's symbols and map all
+// of the symbol references in their contained expressions and interfaces
+// to the new symbols.
+
+struct SymbolAndTypeMappings {
+  std::map<const Symbol *, const Symbol *> symbolMap;
+  std::map<const DeclTypeSpec *, const DeclTypeSpec *> typeMap;
+};
+
+class SymbolMapper : public evaluate::AnyTraverse<SymbolMapper, bool> {
+public:
+  using Base = evaluate::AnyTraverse<SymbolMapper, bool>;
+  SymbolMapper(Scope &scope, SymbolAndTypeMappings &map)
+      : Base{*this}, scope_{scope}, map_{map} {}
+  using Base::operator();
+  bool operator()(const SymbolRef &ref) const {
+    if (const Symbol *mapped{MapSymbol(*ref)}) {
+      const_cast<SymbolRef &>(ref) = *mapped;
+    }
+    return false;
+  }
+  bool operator()(const Symbol &x) const {
+    if (MapSymbol(x)) {
+      DIE("SymbolMapper hit symbol outside SymbolRef");
+    }
+    return false;
+  }
+  void MapSymbolExprs(Symbol &);
+
+private:
+  void MapParamValue(ParamValue &param) const { (*this)(param.GetExplicit()); }
+  void MapBound(Bound &bound) const { (*this)(bound.GetExplicit()); }
+  void MapShapeSpec(ShapeSpec &spec) const {
+    MapBound(spec.lbound());
+    MapBound(spec.ubound());
+  }
+  const Symbol *MapSymbol(const Symbol &) const;
+  const Symbol *MapSymbol(const Symbol *) const;
+  const DeclTypeSpec *MapType(const DeclTypeSpec &);
+  const DeclTypeSpec *MapType(const DeclTypeSpec *);
+  const Symbol *MapInterface(const Symbol *);
+
+  Scope &scope_;
+  SymbolAndTypeMappings &map_;
+};
+
+void SymbolMapper::MapSymbolExprs(Symbol &symbol) {
+  if (auto *object{symbol.detailsIf<ObjectEntityDetails>()}) {
+    if (const DeclTypeSpec *type{object->type()}) {
+      if (const DeclTypeSpec *newType{MapType(*type)}) {
+        object->ReplaceType(*newType);
+      }
+    }
+  }
+  common::visit(
+      common::visitors{[&](ObjectEntityDetails &object) {
+                         for (ShapeSpec &spec : object.shape()) {
+                           MapShapeSpec(spec);
+                         }
+                         for (ShapeSpec &spec : object.coshape()) {
+                           MapShapeSpec(spec);
+                         }
+                       },
+          [&](ProcEntityDetails &proc) {
+            if (const Symbol *
+                mappedSymbol{MapInterface(proc.procInterface())}) {
+              proc.set_procInterface(*mappedSymbol);
+            } else if (const DeclTypeSpec * mappedType{MapType(proc.type())}) {
+              proc.set_type(*mappedType);
+            }
+            if (proc.init()) {
+              if (const Symbol * mapped{MapSymbol(*proc.init())}) {
+                proc.set_init(*mapped);
+              }
+            }
+          },
+          [&](const HostAssocDetails &hostAssoc) {
+            if (const Symbol * mapped{MapSymbol(hostAssoc.symbol())}) {
+              symbol.set_details(HostAssocDetails{*mapped});
+            }
+          },
+          [](const auto &) {}},
+      symbol.details());
+}
+
+const Symbol *SymbolMapper::MapSymbol(const Symbol &symbol) const {
+  if (auto iter{map_.symbolMap.find(&symbol)}; iter != map_.symbolMap.end()) {
+    return iter->second;
+  }
+  return nullptr;
+}
+
+const Symbol *SymbolMapper::MapSymbol(const Symbol *symbol) const {
+  return symbol ? MapSymbol(*symbol) : nullptr;
+}
+
+const DeclTypeSpec *SymbolMapper::MapType(const DeclTypeSpec &type) {
+  if (auto iter{map_.typeMap.find(&type)}; iter != map_.typeMap.end()) {
+    return iter->second;
+  }
+  const DeclTypeSpec *newType{nullptr};
+  if (type.category() == DeclTypeSpec::Category::Character) {
+    const CharacterTypeSpec &charType{type.characterTypeSpec()};
+    if (charType.length().GetExplicit()) {
+      ParamValue newLen{charType.length()};
+      (*this)(newLen.GetExplicit());
+      newType = &scope_.MakeCharacterType(
+          std::move(newLen), KindExpr{charType.kind()});
+    }
+  } else if (const DerivedTypeSpec *derived{type.AsDerived()}) {
+    if (!derived->parameters().empty()) {
+      DerivedTypeSpec newDerived{derived->name(), derived->typeSymbol()};
+      newDerived.CookParameters(scope_.context().foldingContext());
+      for (const auto &[paramName, paramValue] : derived->parameters()) {
+        ParamValue newParamValue{paramValue};
+        MapParamValue(newParamValue);
+        newDerived.AddParamValue(paramName, std::move(newParamValue));
+      }
+      // Scope::InstantiateDerivedTypes() instantiates it later.
+      newType = &scope_.MakeDerivedType(type.category(), std::move(newDerived));
+    }
+  }
+  if (newType) {
+    map_.typeMap[&type] = newType;
+  }
+  return newType;
+}
+
+const DeclTypeSpec *SymbolMapper::MapType(const DeclTypeSpec *type) {
+  return type ? MapType(*type) : nullptr;
+}
+
+const Symbol *SymbolMapper::MapInterface(const Symbol *interface) {
+  if (const Symbol *mapped{MapSymbol(interface)}) {
+    return mapped;
+  }
+  if (interface) {
+    if (&interface->owner() != &scope_) {
+      return interface;
+    } else if (const auto *subp{interface->detailsIf<SubprogramDetails>()};
+               subp && subp->isInterface()) {
+      if (Symbol *newSymbol{scope_.CopySymbol(*interface)}) {
+        newSymbol->get<SubprogramDetails>().set_isInterface(true);
+        map_.symbolMap[interface] = newSymbol;
+        Scope &newScope{scope_.MakeScope(Scope::Kind::Subprogram, newSymbol)};
+        MapSubprogramToNewSymbols(*interface, *newSymbol, newScope, &map_);
+        return newSymbol;
+      }
+    }
+  }
+  return nullptr;
+}
+
+void MapSubprogramToNewSymbols(const Symbol &oldSymbol, Symbol &newSymbol,
+    Scope &newScope, SymbolAndTypeMappings *mappings) {
+  SymbolAndTypeMappings newMappings;
+  if (!mappings) {
+    mappings = &newMappings;
+  }
+  mappings->symbolMap[&oldSymbol] = &newSymbol;
+  const auto &oldDetails{oldSymbol.get<SubprogramDetails>()};
+  auto &newDetails{newSymbol.get<SubprogramDetails>()};
+  for (const Symbol *dummyArg : oldDetails.dummyArgs()) {
+    if (!dummyArg) {
+      newDetails.add_alternateReturn();
+    } else if (Symbol *copy{newScope.CopySymbol(*dummyArg)}) {
+      newDetails.add_dummyArg(*copy);
+      mappings->symbolMap[dummyArg] = copy;
+    }
+  }
+  if (oldDetails.isFunction()) {
+    newScope.erase(newSymbol.name());
+    if (Symbol *copy{newScope.CopySymbol(oldDetails.result())}) {
+      newDetails.set_result(*copy);
+      mappings->symbolMap[&oldDetails.result()] = copy;
+    }
+  }
+  SymbolMapper mapper{newScope, *mappings};
+  for (auto &[_, ref] : newScope) {
+    mapper.MapSymbolExprs(*ref);
+  }
+  newScope.InstantiateDerivedTypes();
+}
+
 } // namespace Fortran::semantics
